package com.marvinelsen.willow

import com.marvinelsen.willow.domain.SearchMode
import com.marvinelsen.willow.ui.DictionaryEntryFx
import com.marvinelsen.willow.ui.SentenceFx
import com.marvinelsen.willow.ui.services.FindCharacterService
import com.marvinelsen.willow.ui.services.FindSentencesService
import com.marvinelsen.willow.ui.services.FindWordsBeginningService
import com.marvinelsen.willow.ui.services.FindWordsContainingService
import com.marvinelsen.willow.ui.services.SearchService
import com.marvinelsen.willow.ui.util.ClipboardHelper
import javafx.beans.property.ObjectProperty
import javafx.beans.property.ReadOnlyBooleanProperty
import javafx.beans.property.ReadOnlyObjectProperty
import javafx.beans.property.SimpleObjectProperty
import javafx.collections.FXCollections
import javafx.collections.ObservableList
import javafx.event.EventHandler

class Model(
    private val searchService: SearchService,
    private val findWordsBeginningService: FindWordsBeginningService,
    private val findWordsContainingService: FindWordsContainingService,
    private val findCharacterService: FindCharacterService,
    private val findSentencesService: FindSentencesService,
) {
    private val internalSelectedEntry: ObjectProperty<DictionaryEntryFx> = SimpleObjectProperty()
    private val internalSearchResults: ObservableList<DictionaryEntryFx> = FXCollections.observableArrayList()
    private val internalWordsBeginning: ObservableList<DictionaryEntryFx> = FXCollections.observableArrayList()
    private val internalWordsContaining: ObservableList<DictionaryEntryFx> = FXCollections.observableArrayList()
    private val internalCharacters: ObservableList<DictionaryEntryFx> = FXCollections.observableArrayList()
    private val internalSentences: ObservableList<SentenceFx> = FXCollections.observableArrayList()

    val selectedEntry: ReadOnlyObjectProperty<DictionaryEntryFx> = internalSelectedEntry

    val searchResults: ObservableList<DictionaryEntryFx> =
        FXCollections.unmodifiableObservableList(internalSearchResults)
    val wordsBeginning: ObservableList<DictionaryEntryFx> =
        FXCollections.unmodifiableObservableList(internalWordsBeginning)
    val wordsContaining: ObservableList<DictionaryEntryFx> =
        FXCollections.unmodifiableObservableList(internalWordsContaining)
    val characters: ObservableList<DictionaryEntryFx> =
        FXCollections.unmodifiableObservableList(internalCharacters)
    val sentences: ObservableList<SentenceFx> =
        FXCollections.unmodifiableObservableList(internalSentences)

    val isSearching: ReadOnlyBooleanProperty = searchService.runningProperty()
    val isFindingWordsBeginning: ReadOnlyBooleanProperty = findWordsBeginningService.runningProperty()
    val isFindingWordsContaining: ReadOnlyBooleanProperty = findWordsContainingService.runningProperty()
    val isFindingCharacters: ReadOnlyBooleanProperty = findCharacterService.runningProperty()
    val isFindingSentences: ReadOnlyBooleanProperty = findSentencesService.runningProperty()

    init {
        searchService.onSucceeded = EventHandler {
            internalSearchResults.setAll(searchService.value)
        }
        findWordsBeginningService.onSucceeded = EventHandler {
            internalWordsBeginning.setAll(findWordsBeginningService.value)
        }
        findWordsContainingService.onSucceeded = EventHandler {
            internalWordsContaining.setAll(findWordsContainingService.value)
        }
        findCharacterService.onSucceeded = EventHandler {
            internalCharacters.setAll(findCharacterService.value)
        }
        findSentencesService.onSucceeded = EventHandler {
            internalSentences.setAll(findSentencesService.value)
        }
    }

    fun search(query: String, searchMode: SearchMode) {
        searchService.searchQuery = query
        searchService.searchMode = searchMode
        searchService.restart()
    }

    fun findWordsBeginning() {
        findWordsBeginningService.entry = internalSelectedEntry.value
        findWordsBeginningService.restart()
    }

    fun findWordsContaining() {
        findWordsContainingService.entry = internalSelectedEntry.value
        findWordsContainingService.restart()
    }

    fun findCharacters() {
        findCharacterService.entry = internalSelectedEntry.value
        findCharacterService.restart()
    }

    fun findSentences() {
        findSentencesService.entry = internalSelectedEntry.value
        findSentencesService.restart()
    }

    fun selectEntry(entry: DictionaryEntryFx) {
        internalWordsBeginning.setAll(emptyList())
        internalWordsContaining.setAll(emptyList())
        internalCharacters.setAll(emptyList())
        internalSentences.setAll(emptyList())
        internalSelectedEntry.value = entry
    }

    fun copyHeadwordOfSelectedEntry() {
        ClipboardHelper.copyString(internalSelectedEntry.value.traditionalProperty.value)
    }

    fun copyPronunciationOfSelectedEntry() {
        ClipboardHelper.copyString(internalSelectedEntry.value.pinyinWithToneMarksProperty.value)
    }
}
